/* ========================================
 *
 * Copyright CONSULTRON, 2020
 * All Rights Reserved
 * UNPUBLISHED, LICENSED SOFTWARE.
 *
 * CONSULTRON allows PUBLIC use of the file.
 * CONSULTRON provides NO WARRANTY expressed or implied.
 *
 * This file is the functions that parse the menu commands.  This is 'generic' and could be common across projects.
 * ========================================
*/
/*
Module: `$INSTANCE_NAME`.c
Module Build: `$CY_MAJOR_VERSION`.`$CY_MINOR_VERSION`.5
Module Comment: `$CY_COMMENT`
•	3.3.0 - Protected help levels are hidden until the protection is unlocked.
•	3.3.3 - Converted Library code into Component.
•	3.3.4 - Removed `$INSTANCE_NAME`_SetMenuCmdArray().  The menu_cmd array is now passed in `$INSTANCE_NAME`_process
•	3.3.5 - Fix: protection code.
*/

#include "`$INSTANCE_NAME`.h"
#include <stdio.h>

#if defined(FREERTOS)
smem_t *scratch_mem = pvTaskGetThreadLocalStoragePointer(NULL,0);	// Get the current task scratch memory
char *tstr = scratch_mem->pSMem;
uint16_t tstr_sz = scratch_mem->SMem_sz;
#else
#define TSTR_SZ 100
char tstr[TSTR_SZ];
#endif

// PW lock switches (if used)
uint8_t pw_locks = 0xFF;		// PW lock switch; TRUE = locked (default)

void PutString_def(char *dummy) { dummy=dummy; return; }		// default function if not set by the user.  (To prevent CPU crash).
const menu_cmd mc_lvl1_def[] = {MENU_END_ALLOC};	// default if not set by the user.  (To prevent CPU crash).

// callback function 
void (*PutString_callback)(char *string) = &PutString_def;	// callback pointer to the PutString()
const menu_cmd *menu_array_Lvl1 = mc_lvl1_def;		// pointer to the Level 1 menucmd array


/*--------------------------------------------------------------------
int `$INSTANCE_NAME`_MenuCmds_Term_SetCallbacks(void (*putstring_cb)(char *string))

This is a function to register a callback function at run-time.
This is used in MenuCmds_process() and MenuCmd_help(). 
This MUST BE set with valid pointers before using the commands mentioned above otherwise:
- NO terminal display will occur.
- No menu commands will be processed.

	Parameters:
		putstring_cb = callback pointer to the Terminal PutString() function.

	Returns:
		0 = No error (For now)
*/
int `$INSTANCE_NAME`_SetCallbacks(void (*putstring_cb)(char *string))
{
	PutString_callback = putstring_cb;	// set the run0time callback
	return(0);
}


// callback function 
void (*num_func_callback)(char c) = NULL;
/*--------------------------------------------------------------------
int `$INSTANCE_NAME`_num_func_SetCallback(void (*num_func_cb))

This is a function to register a callback function at run-time.
This is used in `$INSTANCE_NAME`_process().  
If not set, no special provision is made for processing numbers.

	Parameters:
		cb_ptr = callback pointer to number function to be used.

	Returns:
		0 = No error (For now)
*/
int `$INSTANCE_NAME`_num_func_SetCallback(void (*num_func_cb))
{
	num_func_callback = num_func_cb;	// set the run0time callback
	return(0);	// TODO: support error codes if needed.
}

/*--------------------------------------------------------------------
const menu_cmd *`$INSTANCE_NAME`_search(const menu_cmd menu_array[], uint8_t c)

This function searches the menu_cmd array given for a matching input c.
It will recursively look through the menu_cmd hierarchy until found.

	Parameters:
		menu_array = array pointer to the hierarchy to search.
		c = input byte/char

	Returns:
		*menu_cmd = menu_cmd element FOUND.
		NULL = menu_cmd elementnot FOUND.

Assumptions:
Careful setup of the menu_cmds should be used.  This is a self-recursive function. 
That is why MENU_LEVEL_MAX is used to limit recursiveness.
*/
const menu_cmd *`$INSTANCE_NAME`_search(const menu_cmd menu_array[], uint8_t c)
{
uint8 i=0;
const menu_cmd *menu_elem = NULL;
static uint8 subm_lvl = 0;
_Bool menu_elem_prot = FALSE;
	// Special case where the menu_array is at Level 1
	if(menu_array == menu_array_Lvl1)	subm_lvl = 0;	// If menu_array is at Level 1, then reset the subm_lvl
	if(subm_lvl > MENU_LEVEL_MAX) 
	{	// This is to prevent a SW lockup due to self-recursive pointers.
		return(NULL);	// Exceeds maximum Menu level allowed
	}
	
	for(i=0; menu_array[i].menu_char != eCmdEnd; i++)		// eCmdEnd is the end element
	{
		menu_elem = NULL;
		menu_elem_prot = FALSE;
	 	if((menu_array[i].menu_flags & MENU_PW_MASK) & (pw_locks))	// check any PW locks are active
		{ menu_elem_prot = TRUE; }	// This menu element (function or Help) is currently protected.

		if(c == menu_array[i].menu_char)		// check that menu char matches.
		{	// matched the desired menu character.  
			menu_elem = &menu_array[i];		// pass back the menu element
			break;			// exit for()
		}
		// menu char doesn't match.  Check next level up if this menu element is a "HELP"
		else if((menu_array[i].menu_flags & (MENU_HELP | MENU_HELP_ALWAYS))&& (menu_array[i].menu_func != NULL))
		{ 	// supports a sub menu functions/help
 			if(menu_elem_prot != TRUE) 
			{// PW protected.  Do not pass back the menu element.
				if(menu_array != (menu_cmd *)menu_array[i].menu_func)	// This is to prevent an immediate SW lockup due to self-recursive pointers.
				{
					subm_lvl++;
					menu_elem = `$INSTANCE_NAME`_search( (const menu_cmd *)menu_array[i].menu_func, c); 	// recursive.  menu_elem == NULL if not found or !=  NULL if found.
					subm_lvl--;
				}
				if(menu_elem != NULL) break;  // break out of 'for' loop
			}
		}
	}
	return(menu_elem);
}

/*--------------------------------------------------------------------
void `$INSTANCE_NAME`_help(const menu_cmd help[], _Bool helpmax)

This function processes the menu element in the array that is configured as 'HELP'
	Parameters:
	help = array pointer to the mencmd configured as help.
	helpmax = TRUE = Display all nested levels
			FALSE =  Display only this level

 Returns:
	None

Assumptions:
	For terminal display to work, MenuCmds_PutString_callback must be assigned.
*/
void `$INSTANCE_NAME`_help(const menu_cmd help[], _Bool helpmax)
{
#if defined(FREERTOS)
smem_t *scratch_mem = pvTaskGetThreadLocalStoragePointer(NULL,0);	// Get the current task scratch memory
char *tstr = scratch_mem->pSMem;
uint16_t tstr_sz = scratch_mem->SMem_sz;
#else
uint16_t tstr_sz = sizeof(tstr);
#endif
uint8 i;
static uint8 subm_lvl = 1;
const menu_cmd *subm = (const menu_cmd *)(help->menu_func);

// Special case if at the root menu (L1)
	if(	((help->menu_char == eCmdHelpMin) || helpmax )	&& (subm == NULL)) 
	{	subm = menu_array_Lvl1; subm_lvl = 1; }	// This is a L1 (root) menu help request.  Reset the submenu level
	
	if((subm == NULL)  	// SubMenu pointer is not valid.
	|| (subm_lvl > MENU_LEVEL_MAX)	// Exceeds maximum sMenu level allowed
	) 	return;	// No submenu pointer.  Exit early.
	
	for(i=0; subm[i].menu_char != eCmdEnd; i++)
	{
	_Bool prot_stat = (subm[i].menu_flags & MENU_PW_MASK) & (pw_locks);
		if( (subm[i].menu_flags & MENU_HIDDEN)	 		// if hidden
		) 	{}	// do not display.  This cmd is either currently protected or hidden.
		else
		{	// Cmd not currently protected or hidden.
		char stab[MENU_LEVEL_MAX+2];
		uint8 j;
			for(j=0; j<subm_lvl; j++)
			{
				stab[j] = '\t';
			}
			stab[j] = '\0';	// terminate the stab string
			snprintf(tstr, tstr_sz, "%s%c=> %s%s\r\n", stab, subm[i].menu_char, prot_stat ? "[P] " : "", subm[i].menu_text);
			PutString_callback(tstr);

			if( (helpmax)
			|| (subm[i].menu_flags & MENU_HELP_ALWAYS)	// special case to allow the this commands to be always displayed in help
			)
			{
				if( (prot_stat == 0)	// don't go further if this level is protected.
				&& (subm[i].menu_flags & (MENU_HELP | MENU_HELP_ALWAYS)) // This menu element is a "HELP"
				&& (subm[i].menu_func != NULL)	// subm_func exists
				&& ((const menu_cmd *)(subm[i].menu_func) != (const menu_cmd *)help->menu_func)	// special case:  Not pointing to itself.  Keep from infinite recursiveness
				)
				{
					subm_lvl++;
					`$INSTANCE_NAME`_help((const menu_cmd *)(&subm[i]), helpmax);
					subm_lvl--;
				}
			}
		}
	}
}

/*****************************************************************
char `$INSTANCE_NAME`_process(const menu_cmd menu_array[], uint8_t c)

This function processes the one byte command character.
It determines if a match with one of the menucmd hierarchy exists and 
executes the function or the next level of menu.
	Parameters:
	menu_array = array pointer to the hierarchy to search.
	c = input byte/char

 Returns:
	TRUE = menu command found
	FALSE = menu command not found

Assumptions:
	For terminal display to work, MenuCmds_PutString_callback must be assigned.
*/
_Bool `$INSTANCE_NAME`_process(const menu_cmd menu_array[], uint8_t c)
{
const menu_cmd *menu_elem = NULL;
	menu_array_Lvl1 = menu_array;  // Set the Level 1 array to the input menu_array
// Use only the terminal for protocol changes
	if(c != NO_RX_AVAIL)		// PDL UART =>  CY_SCB_UART_RX_NO_DATA
	{
		switch(c)
		{
// If the callback is defined, we process 0 thru 9 differently, otherwise it's processed like all other cmds.
// This is a run-time feature.
		case '0'...'9':
			if(*num_func_callback != NULL) 
			{
				num_func_callback(c); 
				break;
			}
		default:
			menu_elem = `$INSTANCE_NAME`_search( menu_array_Lvl1, c); 	// recursive.  menu_elem == NULL if not found or !=  NULL if found.
			if(menu_elem != NULL)
			{	// Found the menu char.  Execute Menu function
			_Bool prot_stat = (menu_elem->menu_flags & MENU_PW_MASK) & (pw_locks);

				if(menu_elem->menu_flags & (MENU_HELP | MENU_HELP_ALWAYS)) 
				{	// This menu element is a "HELP"
					snprintf(tstr, sizeof(tstr), "%c => %s%s\r\n", c, prot_stat ? "[P] " : "", menu_elem->menu_text);
					PutString_callback(tstr);
					if(prot_stat == 0)
					{	// Not PW protected.  Continue to parse the menu.
						`$INSTANCE_NAME`_help(menu_elem, (c == eCmdHelpMax ? TRUE : FALSE));
					}
				}
				else 
				{	// This is an executable Command (cmd)
					if(prot_stat == 0)
					{
						snprintf(tstr, sizeof(tstr), "%c ===> %s\r\n",c, menu_elem->menu_text);
						PutString_callback(tstr);
						menu_elem->menu_func(c);
					}
				}
			}
			break;
		}
	}
	return(menu_elem == NULL ? FALSE : TRUE);
}
/* [] END OF FILE */
