/* ========================================
 *
 * Copyright CONSULTRON, 2020
 * All Rights Reserved
 * UNPUBLISHED, LICENSED SOFTWARE.
 *
 * CONSULTRON allows PUBLIC use of the file.
 * CONSULTRON provides NO WARRANTY expressed or implied.
 *
 * ========================================
Module: `$INSTANCE_NAME`.c
Module Build: `$CY_MAJOR_VERSION`.`$CY_MINOR_VERSION`.5
Module Comment: `$CY_COMMENT`

•	1.0.0 - Initial code.   
**  It also uses the CY8CKIT-059 USB port for UART comm instead of the KitProg.
•	1.1.0 - Fix: UART_PutCRLF().  If '\0' sent, we don't add it before the CR and the LF.
	- Add: New function to prevent output data lockups: USBUART_IsReady().
		It tests for USBUART_CheckActivity() when outputing data to prevent a lockup condition if the USB is unplugged.  
		Note: Data attempted to be sent is thrown away just like it would be on a true UART port.
	- Add:  UART_PutArray() compatible function.
•	1.2.0 - Fixed: Update to UART_USB.c to prevent the USB Tx comm from lockup when the host port is not open or the USB cable is unplugged.
	The appears to be no detectable USB signal or interrupt when the USB is unplugged or if the host port is closed.
•	2.0.0 - Convert: Code rewritten as a configurable component.  
		- Configuration UART - Simple UART currently only at 115.2Kbps 8N1. No HW or SW flow control.
		- Configuration USB - USB_FS UART.
•	2.0.1 - Comment only updates
•	2.1.0 - Removed UART v2_50 from the component schematic (.cysch)
	- Add: Incorporated the API files for UART v2_50.  This allows for better UART API call support.
	- Add: Incorporated the symbol file (.cysym) for UART v2_50.  This allows for better HW support.
	- Renamed: Some USBUART.c .h calls to replacement UART API calls.  That simplified it.
•	2.2.1 - Changed: Some parameter category assignments.
•	2.2.2 - Fixed: Issue with aborted PutString().
•	2.2.3 - Added: String_Funcs and MenuCmd callbacks into _Start().
•	2.2.4 - Added: User selectable MenuCmds and String_Funcs components
•	2.2.5 - Fix: Calls to USBUART functions in `$INSTANCE_NAME`.c and `$INSTANCE_NAME`.h
*/
/****************************************************************
PURPOSE:
This "Term" component is intended to allow the user a very simple interface to the a communication interface (UART or USB UART) that shares
the same API calls.   

The designer can choose to select the "UART type" to "UART" or "USB UART".  The UART is available on ALL PSoCs and the USB UART is available on ALL
PSoC5s.

The API calls are a subset of commonly used UART API calls.  The intent is to allow the designer to code for the UART and this component 
easily converts these API calls to the USB UART if selected.  The set of UART API equivalent can be extended in the future if needed.

To use:
- Drop the "Term" component onto a schematic page.
- Configure the component(by double-clicking).  
	- Select either "UART" or "USB UART" as the UART type.
	- Change the component name (if desired).  I commonly change it to "Term".
- If you chose "UART" as the UART type, the comm settings are 115.2Kbps 8N1.  This is hardcoded for now.  You are ready to use it in your App.
- If you chose "USB UART" as the UART type, you need to make the following changes:
	- The USB clock sources need to be set.  Go to the Design Wide Resources\Clocks tab:
		- Set the IMO to "Osc" at "24MHz +/-0.25%"
		- Select the USB clock and chose "IMOx2 - 48MHz"
		- Select the ILO clock and select "100KHz" with Route: "100KHz"
	- Edit the cyapicallbacks.h file.  Add the line:
		#include <`$INSTANCE_NAME`_cyapicallbacks.h>		// include other component callbacks.
- After a successful "Generate Application", you should be able to complete your Application using the UART-equivalent calls.
- If you're project is successfully built and running and you configured as a USB UART, 
	you may need to have your computer recognise the new USB as a COM device.
	In the directory of your project search for a file named "USBUART_cdc.inf".  
	Use this file as the COM device installation file.
- Once the COM device is successfuly assigned, you need to select this COM port using a Terminal program with 115.2Kbps 8N1 as the
	comm parameters.
- If the component is configured for USB UART, you can select the UART at a later time and visa versa.

Implementation limitations at this time:  
The comm API calls to the are mostly blocking.  If needed, a non-blocking means can be provided.
However this requires good structure in the App code to execute this properly (such as data transfer buffer management).
*/

/* ========================================
 * 
 * USBUART ISRs
 * - SOF => Heartbeat @ 1ms.   This interrupt is available only if enabled.
 * - EP0 (Supported) => USB configuration comm.  Occurs when the port is opened at the host.  Used in place of USBUART_GetConfiguration()
 * - EP1 => unknown at this time.
 * - EP2 => Occurs when data to the host is sent
 * - EP3 (Supported) => Occurs when data from the host is available
 * - DP => unknown at this time.
 * - ARB => unknown at this time.
 * - BUS_RESET => Occurs with the USB device is enabled (or plugged in)
 *  At this time only EP0 and EP3 are supported.  Others can be supported by
 *  adding the approriate SW and enabling it in the `$INSTANCE_NAME`_cyapicallbacks.h file.
 * ========================================
*/
#include <project.h>
#include <`$INSTANCE_NAME`.h>

/*--------------------------------------------------------------------------------*/
void `$INSTANCE_NAME`_Start(void)
{
	#if defined(CY_UART_`$INSTANCE_NAME`_UART_H)	// Only allow the defines and prototypes below if UART.h is not defined.
	`$INSTANCE_NAME`_UART_Start();
	#endif
	#if defined(CY_USBFS_`$INSTANCE_NAME`_USBUART_H)	// Only allow the defines and prototypes below if UART.h is not defined.
	`$INSTANCE_NAME`_USBUART_Start(0, `$INSTANCE_NAME`_USBUART_5V_OPERATION);
	#endif
	
	// Add the callbacks for these additional `$INSTANCE_NAME` support functions
#if ENABLE_STRING_FUNCS
	`$INSTANCE_NAME`_String_Funcs_SetCallbacks((void *)&`$INSTANCE_NAME`_GetChar, (void *)&`$INSTANCE_NAME`_PutChar);
#endif
#if ENABLE_MENUCMDS
	`$INSTANCE_NAME`_MenuCmds_SetCallbacks((void *)&`$INSTANCE_NAME`_PutString);
#endif
}

#if defined(CY_USBFS_`$INSTANCE_NAME`_USBUART_H)	// Only allow the defines and prototypes below if UART.h is not defined.

uint8 usb_data[64];
uint16 usb_data_cnt;
_Bool usb_active = `$INSTANCE_NAME`_USBUART_TRUE;
/*--------------------------------------------------------------------------------*/
//_Bool USBUART_IsTxReady(_Bool allow_blocking)
// This prevents a lockup condition if the USB is unplugged or if the host-side has a closed port.
// The SysTick Timer must be enabled to detect that the USB is unplugged or the host port is not open.
//
//	Parameters:
//	allow_blocking = Boolean to determine if the function waits (blocking) or doesn't (non-blocking).
// This can be used as a blocking test function or a non-blocking implementation.  Non-blocking is useful for RTOSes
//
// Returns:
//			TRUE = port ready to accept new data.
//			FALSE = port unplugged or failed.  Do not send new data.
//---------------------------------------------------------
_Bool USBUART_IsTxReady(_Bool allow_blocking)
{
_Bool ret = `$INSTANCE_NAME`_USBUART_TRUE;
	
	if(allow_blocking == `$INSTANCE_NAME`_USBUART_TRUE)
	{	// blocking implementation
	int8 subsystick_wrap_cnt = 0;
	int32_t subsystick_reload = (int32_t)CySysTickGetReload();	// SysTick Reload value
	int32_t subsystick_cnt = (int32_t)CySysTickGetValue();	// SysTick Timestamp when we entered this function
	
		CySysTickStart();		// Start the SysTick interval @ 1ms.  This allows USBUART_IsTxReady() to use the SysTick count for determining if the host is not responding.

		while(`$INSTANCE_NAME`_USBUART_CDCIsReady() == 0)
		{ // Tx Port is not ready.
			if(usb_active == `$INSTANCE_NAME`_USBUART_FALSE) {ret = `$INSTANCE_NAME`_USBUART_FALSE; break;}	// break out of while() because we detected the USB port is not active.
			else
			{	// check the SysTick counter timestamp exceeds up to 2 times the SysTick Reload value (approx. 2ms)
			int32_t subsystick_cnt_diff = subsystick_cnt - (int32)CySysTickGetValue();	// get the difference between the current and prev timestamps
				if(subsystick_cnt_diff < 0) { subsystick_wrap_cnt++; }	// if current subsystick value > prev value increment wrap cnt
				if(( (subsystick_wrap_cnt*subsystick_reload) + (subsystick_cnt_diff)) > (subsystick_reload*2))
				{
					ret = `$INSTANCE_NAME`_USBUART_FALSE;	// Exceeded SysTick 1ms
					break;					// break out of while() because we detected the USB port is not active.
				}
				subsystick_cnt = (int32_t)CySysTickGetValue();							// store new current value
			}
		}

		usb_active = ret;
	}
	else ret = (`$INSTANCE_NAME`_USBUART_CDCIsReady() ? `$INSTANCE_NAME`_USBUART_TRUE : `$INSTANCE_NAME`_USBUART_FALSE);		// non-blocking
	return ret;
}


/*--------------------------------------------------------------------------------*/
// This is a blocking function
void `$INSTANCE_NAME`_USBUART__PutString(const char8 string[]) 
{
	if(USBUART_IsTxReady(`$INSTANCE_NAME`_USBUART_TRUE) == `$INSTANCE_NAME`_USBUART_TRUE) `$INSTANCE_NAME`_USBUART_PutString(string);
}

/*--------------------------------------------------------------------------------*/
// This is a blocking function
void `$INSTANCE_NAME`_USBUART__PutArray(char8 string[], uint16 byteCount)
{
	if(USBUART_IsTxReady(`$INSTANCE_NAME`_USBUART_TRUE) == `$INSTANCE_NAME`_USBUART_TRUE) `$INSTANCE_NAME`_USBUART_PutData((uint8 *)string, byteCount);
}

/*--------------------------------------------------------------------------------*/
// This is a blocking function
void `$INSTANCE_NAME`_USBUART__PutChar(uint8 c)
{
	if(USBUART_IsTxReady(`$INSTANCE_NAME`_USBUART_TRUE) == `$INSTANCE_NAME`_USBUART_TRUE) `$INSTANCE_NAME`_USBUART_PutChar(c);
}

/*--------------------------------------------------------------------------------*/
// This is a blocking function
void `$INSTANCE_NAME`_USBUART__PutCRLF(uint8 c)
{
	if(c != '\0') `$INSTANCE_NAME`_USBUART__PutChar(c);
	`$INSTANCE_NAME`_USBUART__PutString("\r\n");
}

/*--------------------------------------------------------------------------------*/
void `$INSTANCE_NAME`_USBUART__ClearRxBuffer()
{
	usb_data_cnt = 0;
}

/*--------------------------------------------------------------------------------*/
uint8 `$INSTANCE_NAME`_USBUART__ReadRxData()
{
uint8 c = 0;
//	USBUART_check_CDC_init();
	if(	usb_data_cnt >  0)
	{
		c = usb_data[0];
		usb_data_cnt = 0;
	}
	return(c);
}

/*--------------------------------------------------------------------------------*/
uint8 `$INSTANCE_NAME`_USBUART__GetChar()
{
uint8 c = 0;
	c = `$INSTANCE_NAME`_USBUART__ReadRxData();
	return(c);
}

/*--------------------------------------------------------------------------------*/
uint8 `$INSTANCE_NAME`_USBUART__GetRxBufferSize()
{
uint8 count = usb_data_cnt;
	return(count);
}

/*################################################################################################
USBUART ISR support
*/
/*--------------------------------------------------------------------------------*/
#if defined(`$INSTANCE_NAME`_USBUART_EP_0_ISR_EXIT_CALLBACK)
/*--------------------------------------------------------------------------------*/
void `$INSTANCE_NAME`_USBUART_EP_0_ISR_ExitCallback()
{
	PIN_USB_EP0_ISR_Write(1);
	`$INSTANCE_NAME`_USBUART_CDC_Init();
	usb_active = `$INSTANCE_NAME`_USBUART_TRUE;
	PIN_USB_EP0_ISR_Write(0);
}
#endif
// The `$INSTANCE_NAME`_USBUART_EP_0_ISR_ExitCallback() is used inplace of the commented code below. 
///*--------------------------------------------------------------------------------*/
//uint8 `$INSTANCE_NAME`_USBUART_check_CDC_init()
//{
//uint8 ret = USBUART_FAILURE;
//
//    /* Initialize IN endpoints when device is configured. */
//    if (0u != USBUART_GetConfiguration())
//    {
//        /* Enumeration is done, enable OUT endpoint to receive data 
//         * from host. */
//        ret = USBUART_CDC_Init();
//    }
//	return(ret);
//}

#if defined(`$INSTANCE_NAME`_USBUART_EP_1_ISR_EXIT_CALLBACK)
/*--------------------------------------------------------------------------------*/
void `$INSTANCE_NAME`_USBUART_EP_1_ISR_ExitCallback()
{
	PIN_USB_EP1_ISR_Write(1);
	PIN_USB_EP1_ISR_Write(0);
}
#endif

#if defined(`$INSTANCE_NAME`_USBUART_EP_2_ISR_EXIT_CALLBACK)
/*--------------------------------------------------------------------------------*/
void `$INSTANCE_NAME`_USBUART_EP_2_ISR_ExitCallback()
{
	PIN_USB_EP2_ISR_Write(1);
	PIN_USB_EP2_ISR_Write(0);
}
#endif

#if defined(`$INSTANCE_NAME`_USBUART_EP_3_ISR_EXIT_CALLBACK)
/*--------------------------------------------------------------------------------*/
/* ISR that supports incoming data from the host.
	At this time, I am only supporting a single incoming data event. 
	TODO: Support of a FIFO.
*/
void `$INSTANCE_NAME`_USBUART_EP_3_ISR_ExitCallback()
{
	PIN_USB_EP3_ISR_Write(1);
	if(`$INSTANCE_NAME`_USBUART_GetCount() > 0)
	{
		usb_data_cnt = `$INSTANCE_NAME`_USBUART_GetAll(usb_data);	// Get all incoming data
	}
	PIN_USB_EP3_ISR_Write(0);
}
#endif

#if defined(`$INSTANCE_NAME`_USBUART_ARB_ISR_EXIT_CALLBACK)
/*--------------------------------------------------------------------------------*/
void `$INSTANCE_NAME`_USBUART_ARB_ISR_ExitCallback()
{
	PIN_USB_ARB_ISR_Write(1);
	PIN_USB_ARB_ISR_Write(0);
}
#endif

#if defined(`$INSTANCE_NAME`_USBUART_BUS_RESET_ISR_EXIT_CALLBACK)
/*--------------------------------------------------------------------------------*/
void `$INSTANCE_NAME`_USBUART_BUS_RESET_ISR_ExitCallback()
{
	PIN_USB_BUS_RESET_ISR_Write(1);
	PIN_USB_BUS_RESET_ISR_Write(0);
}
#endif

#if defined(`$INSTANCE_NAME`_USBUART_DP_ISR_EXIT_CALLBACK)
/*--------------------------------------------------------------------------------*/
void `$INSTANCE_NAME`_USBUART_DP_ISR_ExitCallback()
{
	PIN_USB_DP_ISR_Write(1);
	PIN_USB_DP_ISR_Write(0);
}
#endif

#if defined(`$INSTANCE_NAME`_USBUART_SOF_ISR_EXIT_CALLBACK)
/*--------------------------------------------------------------------------------*/
void `$INSTANCE_NAME`_USBUART_SOF_ISR_ExitCallback()
{
	PIN_USB_SOF_Write(1);
	PIN_USB_SOF_Write(0);
}
#endif

#endif		// #if !defined(CY_UART_`$INSTANCE_NAME`_UART_H)
/* [] END OF FILE */
